/* Percepts provided by the world are as follows:
 *   gsize(N,X,Y) - the current grid is scenario N, dimensions X by Y (cells 0 to (X-1))
 *   depot(N,X,Y) - Grid scenario N, and the gold depot is at X,Y 
 *   pos(X,Y) - the agent's current position
 *   carrying_gold - this belief will be held if the agent has picked up gold.
 *   cell(X,Y,Object) - if there's an object in any of the 8 adjacent cells, then 
	    				these beliefs will hold. Object can be "obstacle", "gold", or "ally".
		    			Note ally will also appear for the current agent, at its position

 * Actions availible are of the form do(Action) where Action is:
 *   up - to move up (Y-1 - note 0,0 is the top left of the grid)
 *   down - to move down (Y+1)
 *   right - to move right (X+1)
 *   left - to move left (X-1)
 *   pick - to pick up some gold, only 1 piece can be carried at one time.
 *   drop - to drop some gold.
 */

/* This is the initial goal.*/

!goTo(18, 4).
//!explore(2, 2).

/* Robot will explore the part of the world containing (X, Y). */

+!explore(X, Y) : true
	<- +corners(0, 0, 20, 20); // remember the upper-left and bottom-right corner
	+innerCorners(1, 1, 19, 19); // remember the upper-left and bottom-right inner corner
	+begin(1, 1); // remember the point where the exploration is going to start
	!goTo(1, 1); // go to the starting position in my part of the world
	!checkForGold; // check if there is gold around starting position
	!exploration. // perform the exploration of the area

/* traverse horizontally my part of the environment */

+!traverseRight : innerCorners(_, _, X, _) & pos(X, _). // base case

+!traverseRight : true // loop case
	<- !moveAndCheck(right);
	!traverseRight.

+!traverseLeft : innerCorners(X, _, _, _) & pos(X, _). // base case

+!traverseLeft : true // loop case
	<- !moveAndCheck(left);
	!traverseLeft.
	
/* perform the exploration of my part of the environment */	

+!exploration : explorationFinished & myPart(P) // exploration is finished
	<- .print("Exploration of part ", P, " finished!").

+!exploration : true // move horizontally, then vertically and then continue with exploration
	<- !moveHorizontally;
	!moveVertically;
	!exploration.
	
+!moveHorizontally : innerCorners(X, _, _, _) & pos(X, _)
	<- !traverseRight.	

+!moveHorizontally : innerCorners(_, _, X, _) & pos(X, _)
	<- !traverseLeft.	
	
// move vertically up
+!moveVertically : begin(_, Ybegin) & innerCorners(_, Y1, _, Y2) & (Ybegin == Y2) & pos(_, Ypos)
	<- !moveVerticallyUp(Ypos, Y1). // Ypos ... curent Y-position, Y1 ... final Y-position

// we are at the final line	
+!moveVerticallyUp(Y, Y1) : (Y == Y1)			<- +explorationFinished.

// we have to move up once or twice to reach the final line
+!moveVerticallyUp(Y, Y1) : ((Y - Y1) == 1)		<- !moveAndCheck(up).
+!moveVerticallyUp(Y, Y1) : ((Y - Y1) == 2)		<- !moveAndCheck(up); !moveAndCheck(up).

// normally, we move by three steps
+!moveVerticallyUp(Y, Y1) : ((Y - Y1) >= 3)		<- !moveAndCheck(up); !moveAndCheck(up); !moveAndCheck(up).

// move vertically down
+!moveVertically : begin(_, Ybegin) & innerCorners(_, Y1, _, Y2) & (Ybegin == Y1) & pos(_, Ypos)
	<- !moveVerticallyDown(Ypos, Y2).
	
// same as for vertical movement up
+!moveVerticallyDown(Y, Y2) : (Y == Y2)			<- +explorationFinished.
+!moveVerticallyDown(Y, Y2) : ((Y2 - Y) == 1)	<- !moveAndCheck(down).
+!moveVerticallyDown(Y, Y2) : ((Y2 - Y) == 2)	<- !moveAndCheck(down); !moveAndCheck(down).
+!moveVerticallyDown(Y, Y2) : ((Y2 - Y) >= 3) 	<- !moveAndCheck(down); !moveAndCheck(down); !moveAndCheck(down).

/* Perform a move and check if there is gold around */

+!moveAndCheck(Direction) : true
	<- //!move(Direction);
	do(Direction);
	!checkForGold.

+!checkForGold : cell(X, Y, gold)
	<- .findall([Xg, Yg], cell(Xg, Yg, gold), L); // all gold around is in the list L
	!announceGold(L).
	
+!checkForGold : true. // there is no gold around

/* If there is gold around, remember its position and announce it */

+!announceGold([]). // base case

+!announceGold(L) : (L = [[X, Y]|T]) & (goldAt(X, Y)) // position of gold is already in KB
	<- !announceGold(T).

+!announceGold(L) : (L = [[X, Y]|T]) & (not (goldAt(X, Y))) // position of gold is not in KB
	<- +goldAt(X, Y); // remember position
	.print("Gold at: [", X, ", ", Y, "]");
	.broadcast(tell, goldAt(X, Y)); // announce the position of gold
	!announceGold(T). // process the rest of the list

/* go routine - robot is going to position (X, Y) */

+!goTo(X, Y) : true // go to the point (X, Y)
	<- +goTo(X, Y); // remember the destination
	!go(X, Y).

+!go(X, Y) : pos(X, Y) // we are at the destination.
	<- -goTo(X, Y). // delete the destination from our beliefs
	//.print("Successfully arrived at (", X, ", ", Y, ")."). 

// move left 
+!go(X, Y) : pos(Xc, Yc) & (Xc > X)
	<- //!move(left);
	!move(left);
	!go(X, Y).
	
// move right 
+!go(X, Y) : pos(Xc, Yc) & (Xc < X) 
	<- //!move(right);
	!move(right);
	!go(X, Y).

// move up
+!go(X, Y) : pos(Xc, Yc) & (Yc > Y)
	<- //!move(up);
	!move(up);
	!go(X,Y).
	
// move down
+!go(X, Y) : pos(Xc, Yc) & (Yc < Y)
	<- //!move(down);
	!move(down);
	!go(X, Y).
	
+!go(X, Y) : pos(Xc, Yc) // error case
	<- .println("ERROR! Trying to go: ", X, Y, " from: ", Xc, Yc).	

/* randomized collision */

// TODO
	
/* no collision avoidance */

+!move(Dir) : true
	<- do(Dir).

